package libsvm
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Module Svm.ProblemSource
Type of a SVM problem (training set).
create x y constructs a problem from a feature matrix x and target vector y. Each row of x is a feature vector of a training instance.
create_k k y constructs a problem from a matrix k and target vector y. The matrix k has to be of the following form:
1 K(x1,x1) K(x1,x2) ... K(x1,xL)
2 K(x2,x1) K(x2,x2) ... K(x2,xL)
...
L K(xL,x1) K(xL,x2) ... K(xL,xL)
where L denotes the number of training instances and K(x,y) is the precomputed kernel value of the two training instances x and y.
get_targets prob
output prob oc outputs the problem prob to an output channel oc. NOTE: the function does not close the output channel.
min_max_feats prob
val scale :
?lower:float ->
?upper:float ->
t ->
min_feats:Lacaml.D.vec ->
max_feats:Lacaml.D.vec ->
tscale ?lower ?upper prob min_feats max_feats